import pytest

import ipaddress

from asn_check import ip_binary_tree


def test_ip_search():
    init = [
        {"ip": ipaddress.IPv4Network("127.0.0.0/8"), "label": "local"},
        {"ip": ipaddress.IPv4Network("192.168.0.0/16"), "label": "home"},
        {"ip": ipaddress.IPv4Network("54.168.1.12/32"), "label": "single"},
    ]
    cases = [
        {
            "label": "localhost",
            "input": {"ip": ipaddress.IPv4Address("127.0.0.1")},
            "output": "local",
            "ex": None,
        },
        {
            "label": "router",
            "input": {"ip": ipaddress.IPv4Address("192.168.0.1")},
            "output": "home",
            "ex": None,
        },
        {
            "label": "not router",
            "input": {"ip": ipaddress.IPv4Address("192.165.0.1")},
            "output": None,
            "ex": None,
        },
        {
            "label": "single",
            "input": {"ip": ipaddress.IPv4Address("54.168.1.12")},
            "output": "single",
            "ex": None,
        },
        {
            "label": "not local",
            "input": {"ip": ipaddress.IPv4Address("34.0.0.15")},
            "output": None,
            "ex": None,
        },
    ]
    it = ip_binary_tree.IPTree()
    for inp in init:
        it.add_ip(**inp)
    for case in cases:
        try:
            output = it.search(**case["input"])
            assert output == case["output"], "case '{}', output: exp {}, got {}".format(
                case["label"], case["output"], output
            )
        except Exception as e:
            assert type(e) == case["ex"], "case '{}', ex: exp {}, got {}".format(case["label"], case["ex"], type(e))



from requests_cache import CachedSession
from platformdirs import user_cache_dir
import ipaddress
from collections import defaultdict
import logging

log = logging.getLogger("__main__")

ASN_ROUTES_URL = 'https://thyme.apnic.net/current/data-raw-table'
ASN_NAMES_URL = 'https://ftp.ripe.net/ripe/asnames/asn.txt'


def get_data():
    session = CachedSession(user_cache_dir('asn_check', 'mh'), cache_control=True, backend='filesystem')
    log.info(f"Getting ASN routes from {ASN_ROUTES_URL}")
    asn_routes = session.get(ASN_ROUTES_URL)
    log.info(f"Getting ASN names from {ASN_NAMES_URL}")
    asn_names = session.get(ASN_NAMES_URL)
    return asn_routes.text, asn_names.text


def parse_asn_routes(asn_routes: str):
    """
    -> dict[str,list[ipaddress.IPv4Network]]
    """
    result = defaultdict(list)
    for line in asn_routes.splitlines():
        try:
            p = line.split('\t')
            result[p[1]].append(ipaddress.IPv4Network(p[0]))
        except:
            log.error(f"Invalid ASN route format: {line}")
    return result


def parse_asn_names(asn_names: str):
    """
    -> dict[str,dict[str,str]]
    """
    result = dict()
    for line in asn_names.splitlines():
        try:
            sp = line.find(' ')
            c = line.find(',')
            asn = line[:sp]
            name = line[sp + 1 : c]
            code = line[c + 2 :]
            result[asn] = {'name': name, 'country_code': code}
        except:
            log.error(f"Invalid ASN name format: {line}")
    return result



import pytest

from asn_check import utils


def test_ipv4_re():
    cases = [
        {
            "label": "localhost",
            "input": "127.0.0.1",
            "output": "127.0.0.1",
            "ex": None,
        },
        {
            "label": "low",
            "input": "0.0.0.0",
            "output": "0.0.0.0",
            "ex": None,
        },
        {
            "label": "high",
            "input": "255.255.255.255",
            "output": "255.255.255.255",
            "ex": None,
        },
        {
            "label": "higher",
            "input": "256.255.255.255",
            "output": None,
            "ex": None,
        },
        {
            "label": "higher 2",
            "input": "456.555.255.255",
            "output": None,
            "ex": None,
        },
        {
            "label": "zero",
            "input": "056.255.255.255",
            "output": None,
            "ex": None,
        },
        {
            "label": "too many digits",
            "input": "2216.255.255.255",
            "output": None,
            "ex": None,
        },
    ]
    for case in cases:
        try:
            output = utils.ipv4_re.match(case["input"])
            if output:
                output = output.group(0)
            assert output == case["output"], "case '{}', output: exp {}, got {}".format(
                case["label"], case["output"], output
            )
        except Exception as e:
            assert type(e) == case["ex"], "case '{}', ex: exp {}, got {}".format(case["label"], case["ex"], type(e))


def test_ipv4_net_re():
    cases = [
        {
            "label": "localhost",
            "input": "127.0.0.1/10",
            "output": "127.0.0.1/10",
            "ex": None,
        },
        {
            "label": "low",
            "input": "0.0.0.0/0",
            "output": "0.0.0.0/0",
            "ex": None,
        },
        {
            "label": "high",
            "input": "255.255.255.255/32",
            "output": "255.255.255.255/32",
            "ex": None,
        },
        {
            "label": "higher",
            "input": "256.255.255.255/32",
            "output": None,
            "ex": None,
        },
        {
            "label": "higher net",
            "input": "255.255.255.255/33",
            "output": None,
            "ex": None,
        },
        {
            "label": "higher 2",
            "input": "456.555.255.255/32",
            "output": None,
            "ex": None,
        },
        {
            "label": "higher net 2",
            "input": "456.555.255.255/62",
            "output": None,
            "ex": None,
        },
        {
            "label": "zero",
            "input": "056.255.255.255/12",
            "output": None,
            "ex": None,
        },
        {
            "label": "zero net",
            "input": "056.255.255.255/03",
            "output": None,
            "ex": None,
        },
        {
            "label": "too many digits",
            "input": "2216.255.255.255/10",
            "output": None,
            "ex": None,
        },
        {
            "label": "too many digits net",
            "input": "2216.255.255.255/123",
            "output": None,
            "ex": None,
        },
    ]
    for case in cases:
        try:
            output = utils.ipv4_net_re.match(case["input"])
            if output:
                output = output.group(0)
            assert output == case["output"], "case '{}', output: exp {}, got {}".format(
                case["label"], case["output"], output
            )
        except Exception as e:
            assert type(e) == case["ex"], "case '{}', ex: exp {}, got {}".format(case["label"], case["ex"], type(e))



import pytest
import ipaddress

from asn_check import data


def test_parse_asn_routes():
    cases = [
        {
            "label": "happy",  # I don't really expect them to break the format
            "input": {
                "asn_routes": """1.0.0.0/24	13335
1.0.4.0/22	38803
1.0.5.0/24	38803
1.0.16.0/24	2519
1.0.128.0/17	23969
1.0.128.0/18	23969
1.0.128.0/19	23969
1.0.128.0/24	23969
1.0.129.0/24	23969
1.0.130.0/23	23969"""
            },
            "output": {
                "13335": [ipaddress.IPv4Network("1.0.0.0/24")],
                "38803": [ipaddress.IPv4Network("1.0.4.0/22"), ipaddress.IPv4Network("1.0.5.0/24")],
                "2519": [ipaddress.IPv4Network("1.0.16.0/24")],
                "23969": [
                    ipaddress.IPv4Network("1.0.128.0/17"),
                    ipaddress.IPv4Network("1.0.128.0/18"),
                    ipaddress.IPv4Network("1.0.128.0/19"),
                    ipaddress.IPv4Network("1.0.128.0/24"),
                    ipaddress.IPv4Network("1.0.129.0/24"),
                    ipaddress.IPv4Network("1.0.130.0/23"),
                ],
            },
            "ex": None,
        },
    ]
    for case in cases:
        try:
            output = data.parse_asn_routes(**case["input"])
            assert output == case["output"], "case '{}', output: exp {}, got {}".format(
                case["label"], case["output"], output
            )
        except Exception as e:
            assert type(e) == case["ex"], "case '{}', ex: exp {}, got {}".format(case["label"], case["ex"], type(e))


def test_parse_asn_names():
    cases = [
        {
            "label": "happy",  # I don't really expect them to break the format
            "input": {
                "asn_names": """25 UCB, US
26 CORNELL, US
27 UMDNET, US
28 DFVLR-SYS Deutsches Zentrum fuer Luft- und Raumfahrt e.V., DE"""
            },
            "output": {
                "25": {"name": "UCB", "country_code": "US"},
                "26": {"name": "CORNELL", "country_code": "US"},
                "27": {"name": "UMDNET", "country_code": "US"},
                "28": {"name": "DFVLR-SYS Deutsches Zentrum fuer Luft- und Raumfahrt e.V.", "country_code": "DE"},
            },
            "ex": None,
        },
    ]
    for case in cases:
        try:
            output = data.parse_asn_names(**case["input"])
            assert output == case["output"], "case '{}', output: exp {}, got {}".format(
                case["label"], case["output"], output
            )
        except Exception as e:
            assert type(e) == case["ex"], "case '{}', ex: exp {}, got {}".format(case["label"], case["ex"], type(e))



#!/usr/bin/env python3

import os
import json
import csv
import logging
import sys
import time
import math
import ipaddress
import shelve
from functools import update_wrapper

import click
import requests

from asn_check.ip_binary_tree import IPTree
from asn_check.data import get_data, parse_asn_routes, parse_asn_names
from asn_check.utils import ipv4_re, ipv4_net_re


logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s [%(levelname)s] %(message)s",
    handlers=[logging.StreamHandler()],
)
log = logging.getLogger("__main__")


log_levels = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
    "CRITICAL": logging.CRITICAL,
}


def log_decorator(f):
    @click.pass_context
    def new_func(ctx, *args, **kwargs):
        log.setLevel(log_levels[ctx.params["log_level"]])
        log.info("Starting")
        r = ctx.invoke(f, *args, **kwargs)
        log.info("Finishing")
        return r

    return update_wrapper(new_func, f)


def time_decorator(f):
    @click.pass_context
    def new_func(ctx, *args, **kwargs):
        t1 = time.perf_counter()
        try:
            r = ctx.invoke(f, *args, **kwargs)
            return r
        except Exception as e:
            raise e
        finally:
            t2 = time.perf_counter()
            mins = math.floor(t2 - t1) // 60
            hours = mins // 60
            secs = (t2 - t1) - 60 * mins - 3600 * hours
            log.info(f"Execution in {hours:02d}:{mins:02d}:{secs:0.4f}")

    return update_wrapper(new_func, f)


@click.command()
@click.option(
    "--input-file",
    help="Input file with one IPv4 per line [default: STDIN]",
    type=click.File("rt"),
    default=sys.stdin,
)
@click.option(
    "--output-file",
    help="Output file - csv, header: ip,asn,name,country_code [default: STDOUT]",
    type=click.File("at"),
    default=sys.stdout,
)
@click.option(
    "--log-level",
    default="WARNING",
    type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]),
    show_default=True,
    help="Set logging level.",
    envvar="LOG_LEVEL",
)
@log_decorator
@time_decorator
def main(input_file, output_file, log_level):
    """Console script for asn_check

    If you have a list of IPv4 addresses, this script can assign AS nums, AS names and country codes to those addresses.

    Queries whois services (with caching).

    Returns a CSV file."""
    # ======================================================================
    #                        Your script starts here!
    # ======================================================================
    log.info(f"Get data")
    asn_routes, asn_names = get_data()
    log.info(f"Parse data")
    all_nets = parse_asn_routes(asn_routes)
    names = parse_asn_names(asn_names)

    log.info(f"Construct the tree")
    iptree = IPTree()
    for asn in all_nets:
        for net in all_nets[asn]:
            iptree.add_ip(net, f"{asn}")

    log.info("Load addresses")
    if input_file.isatty():
        logging.critical("Input from stdin which is a tty - aborting")
        return 128
    with input_file:
        in_data = input_file.read().splitlines()
    addresses = [ipaddress.IPv4Address(addr) for addr in in_data if ipv4_re.match(addr)]
    log.info(f"Got {len(addresses)} addresses")

    log.info(f"Searching...")
    header = ["ip", "asn", "name", "country_code"]
    writer = csv.DictWriter(output_file, fieldnames=header)
    writer.writeheader()
    for addr in addresses:
        label = iptree.search(addr)
        meta = names.get(label, {"name": "", "country_code": ""})
        writer.writerow({"ip": addr, "asn": label, "name": meta["name"], "country_code": meta["country_code"]})
    return 0


if __name__ == "__main__":
    main()



from click.testing import CliRunner
from asn_check.asn_check import main
import pytest


def test_e2e():
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir='/tmp/click_tests/') as td:
        with open('input.txt', 'w') as f:
            f.write(
                """250.254.147.119
12.154.0.67
41.13.122.240
176.218.30.1078
128.105.177.84
85.227.158.196
74.74.207.74"""
            )
        result = runner.invoke(main, ["--input-file", "input.txt", "--output-file", "output.txt"])
        assert result.exit_code == 0
        with open('output.txt', 'r') as f:
            output = f.read()
        assert (
            output
            == """ip,asn,name,country_code
250.254.147.119,,,
12.154.0.67,7018,ATT-INTERNET4,US
41.13.122.240,29975,VODACOM-,ZA
128.105.177.84,59,WISC-MADISON-AS,US
85.227.158.196,2119,TELENOR-NEXTEL Telenor Norge AS,NO
74.74.207.74,11351,TWC-11351-NORTHEAST,US
"""
        )



import ipaddress


class IPNode:
    def __init__(self, value=None):
        self.one = None
        self.zero = None
        self.value = value


class IPTree:
    def __init__(self):
        self.root = IPNode()

    def add_ip(self, ip: ipaddress.IPv4Network, label: str):
        addr_netmask = bin(int(ip.netmask))[2:].rjust(32, "0")
        addr_ip = bin(int(ip.network_address))[2:].rjust(32, "0")
        addr_pointer = 0
        node_pointer = self.root
        for c in addr_netmask:
            if c == "0":
                node_pointer.value = label
                break
            if addr_ip[addr_pointer] == "0":
                if node_pointer.zero is None:
                    node_pointer.zero = IPNode()
                node_pointer = node_pointer.zero
            else:
                if node_pointer.one is None:
                    node_pointer.one = IPNode()
                node_pointer = node_pointer.one
            addr_pointer += 1
            if addr_pointer == 32:
                node_pointer.value = label

    def search(self, ip: ipaddress.IPv4Address) -> str:
        node_pointer = self.root
        addr_ip = bin(int(ip))[2:].rjust(32, "0")
        addr_pointer = 0
        for c in addr_ip:
            if node_pointer is None:
                return None
            if node_pointer.value is not None:
                return node_pointer.value
            if c == "0":
                node_pointer = node_pointer.zero
            else:
                node_pointer = node_pointer.one
            addr_pointer += 1
            if addr_pointer == 32:
                if node_pointer is not None and node_pointer.value is not None:
                    return node_pointer.value
        return None
